@@ -1,7 +1,7 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; frozen_string_literal: true -*-
 #
 #--
-# Copyright (C) 2009-2016 Thomas Leitner <t_leitner@gmx.at>
+# Copyright (C) 2009-2019 Thomas Leitner <t_leitner@gmx.at>
 #
 # This file is part of kramdown which is licensed under the MIT.
 #++
@@ -29,16 +29,18 @@ module Kramdown
         @stack = []
       end
 
-      def convert(el, opts = {:indent => 0})
+      def convert(el, opts = {indent: 0})
         res = send("convert_#{el.type}", el, opts)
+        res = res.dup if res.frozen?
         if ![:html_element, :li, :dt, :dd, :td].include?(el.type) && (ial = ial_for_element(el))
           res << ial
-          res << "\n\n" if Element.category(el) == :block
+          res << "\n\n" if el.block?
         elsif [:ul, :dl, :ol, :codeblock].include?(el.type) && opts[:next] &&
             ([el.type, :codeblock].include?(opts[:next].type) ||
-             (opts[:next].type == :blank && opts[:nnext] && [el.type, :codeblock].include?(opts[:nnext].type)))
+             (opts[:next].type == :blank && opts[:nnext] &&
+              [el.type, :codeblock].include?(opts[:nnext].type)))
           res << "^\n\n"
-        elsif Element.category(el) == :block &&
+        elsif el.block? &&
             ![:li, :dd, :dt, :td, :th, :tr, :thead, :tbody, :tfoot, :blank].include?(el.type) &&
             (el.type != :html_element || @stack.last.type != :html_element) &&
             (el.type != :p || !el.options[:transparent])
@@ -47,23 +49,23 @@ module Kramdown
         res
       end
 
-      def inner(el, opts = {:indent => 0})
+      def inner(el, opts = {indent: 0})
         @stack.push(el)
-        result = ''
+        result = +''
         el.children.each_with_index do |inner_el, index|
           options = opts.dup
           options[:index] = index
-          options[:prev] = (index == 0 ? nil : el.children[index-1])
-          options[:pprev] = (index <= 1 ? nil : el.children[index-2])
-          options[:next] = (index == el.children.length - 1 ? nil : el.children[index+1])
-          options[:nnext] = (index >= el.children.length - 2 ? nil : el.children[index+2])
+          options[:prev] = (index == 0 ? nil : el.children[index - 1])
+          options[:pprev] = (index <= 1 ? nil : el.children[index - 2])
+          options[:next] = (index == el.children.length - 1 ? nil : el.children[index + 1])
+          options[:nnext] = (index >= el.children.length - 2 ? nil : el.children[index + 2])
           result << convert(inner_el, options)
         end
         @stack.pop
         result
       end
 
-      def convert_blank(el, opts)
+      def convert_blank(_el, _opts)
         ""
       end
 
@@ -82,8 +84,8 @@ module Kramdown
       def convert_p(el, opts)
         w = @options[:line_width] - opts[:indent].to_s.to_i
         first, second, *rest = inner(el, opts).strip.gsub(/(.{1,#{w}})( +|$\n?)/, "\\1\n").split(/\n/)
-        first.gsub!(/^(?:(#|>)|(\d+)\.|([+-]\s))/) { $1 || $3 ? "\\#{$1 || $3}" : "#{$2}\\."} if first
-        second.gsub!(/^([=-]+\s*?)$/, "\\\1") if second
+        first&.gsub!(/^(?:(#|>)|(\d+)\.|([+-]\s))/) { $1 || $3 ? "\\#{$1 || $3}" : "#{$2}\\." }
+        second&.gsub!(/^([=-]+\s*?)$/, "\\\1")
         res = [first, second, *rest].compact.join("\n") + "\n"
         if el.children.length == 1 && el.children.first.type == :math
           res = "\\#{res}"
@@ -93,41 +95,40 @@ module Kramdown
         res
       end
 
-
-      def convert_codeblock(el, opts)
-        el.value.split(/\n/).map {|l| l.empty? ? "    " : "    #{l}"}.join("\n") + "\n"
+      def convert_codeblock(el, _opts)
+        el.value.split(/\n/).map {|l| l.empty? ? "    " : "    #{l}" }.join("\n") + "\n"
       end
 
       def convert_blockquote(el, opts)
         opts[:indent] += 2
-        inner(el, opts).chomp.split(/\n/).map {|l| "> #{l}"}.join("\n") << "\n"
+        inner(el, opts).chomp.split(/\n/).map {|l| "> #{l}" }.join("\n") << "\n"
       end
 
       def convert_header(el, opts)
-        res = ''
+        res = +''
         res << "#{'#' * output_header_level(el.options[:level])} #{inner(el, opts)}"
-        res[-1, 1] = "\\#" if res[-1] == ?#
+        res[-1, 1] = "\\#" if res[-1] == '#'
         res << "   {##{el.attr['id']}}" if el.attr['id'] && !el.attr['id'].strip.empty?
         res << "\n"
       end
 
-      def convert_hr(el, opts)
+      def convert_hr(_el, _opts)
         "* * *\n"
       end
 
       def convert_ul(el, opts)
         inner(el, opts).sub(/\n+\Z/, "\n")
       end
-      alias :convert_ol :convert_ul
-      alias :convert_dl :convert_ul
+      alias convert_ol convert_ul
+      alias convert_dl convert_ul
 
       def convert_li(el, opts)
         sym, width = if @stack.last.type == :ul
-                       ['* ', el.children.first && el.children.first.type == :codeblock ? 4 : 2]
+                       [+'* ', el.children.first && el.children.first.type == :codeblock ? 4 : 2]
                      else
                        ["#{opts[:index] + 1}.".ljust(4), 4]
                      end
-        if ial = ial_for_element(el)
+        if (ial = ial_for_element(el))
           sym << ial << " "
         end
 
@@ -135,12 +136,12 @@ module Kramdown
         text = inner(el, opts)
         newlines = text.scan(/\n*\Z/).first
         first, *last = text.split(/\n/)
-        last = last.map {|l| " "*width + l}.join("\n")
+        last = last.map {|l| " " * width + l }.join("\n")
         text = (first.nil? ? "\n" : first + (last.empty? ? "" : "\n") + last + newlines)
         if el.children.first && el.children.first.type == :p && !el.children.first.options[:transparent]
-          res = "#{sym}#{text}"
+          res = +"#{sym}#{text}"
           res << "^\n" if el.children.size == 1 && @stack.last.children.last == el &&
-            (@stack.last.children.any? {|c| c.children.first.type != :p} || @stack.last.children.size == 1)
+            (@stack.last.children.any? {|c| c.children.first.type != :p } || @stack.last.children.size == 1)
           res
         elsif el.children.first && el.children.first.type == :codeblock
           "#{sym}\n    #{text}"
@@ -150,8 +151,8 @@ module Kramdown
       end
 
       def convert_dd(el, opts)
-        sym, width = ": ", (el.children.first && el.children.first.type == :codeblock ? 4 : 2)
-        if ial = ial_for_element(el)
+        sym, width = +": ", (el.children.first && el.children.first.type == :codeblock ? 4 : 2)
+        if (ial = ial_for_element(el))
           sym << ial << " "
         end
 
@@ -159,10 +160,10 @@ module Kramdown
         text = inner(el, opts)
         newlines = text.scan(/\n*\Z/).first
         first, *last = text.split(/\n/)
-        last = last.map {|l| " "*width + l}.join("\n")
+        last = last.map {|l| " " * width + l }.join("\n")
         text = first.to_s + (last.empty? ? "" : "\n") + last + newlines
         text.chomp! if text =~ /\n\n\Z/ && opts[:next] && opts[:next].type == :dd
-        text << "\n" if (text !~ /\n\n\Z/ && opts[:next] && opts[:next].type == :dt)
+        text << "\n" if text !~ /\n\n\Z/ && opts[:next] && opts[:next].type == :dt
         text << "\n" if el.children.empty?
         if el.children.first && el.children.first.type == :p && !el.children.first.options[:transparent]
           "\n#{sym}#{text}"
@@ -174,34 +175,42 @@ module Kramdown
       end
 
       def convert_dt(el, opts)
-        result = ''
-        if ial = ial_for_element(el)
+        result = +''
+        if (ial = ial_for_element(el))
           result << ial << " "
         end
         result << inner(el, opts) << "\n"
       end
 
-      HTML_TAGS_WITH_BODY=['div', 'script', 'iframe', 'textarea']
+      HTML_TAGS_WITH_BODY = ['div', 'script', 'iframe', 'textarea', 'th', 'td']
+
+      HTML_ELEMENT_TYPES = [:entity, :text, :html_element].freeze
+      private_constant :HTML_ELEMENT_TYPES
 
       def convert_html_element(el, opts)
         markdown_attr = el.options[:category] == :block && el.children.any? do |c|
-          c.type != :html_element && (c.type != :p || !c.options[:transparent]) && Element.category(c) == :block
+          c.type != :html_element &&
+            (c.type != :p || !c.options[:transparent] ||
+             c.children.any? {|t| !HTML_ELEMENT_TYPES.member?(t.type) }) &&
+            c.block?
         end
-        opts[:force_raw_text] = true if %w{script pre code}.include?(el.value)
-        opts[:raw_text] = opts[:force_raw_text] || opts[:block_raw_text] || (el.options[:category] != :span && !markdown_attr)
+        opts[:force_raw_text] = true if %w[script pre code].include?(el.value)
+        opts[:raw_text] = opts[:force_raw_text] || opts[:block_raw_text] || \
+          (el.options[:category] != :span && !markdown_attr)
         opts[:block_raw_text] = true if el.options[:category] == :block && opts[:raw_text]
         res = inner(el, opts)
         if el.options[:category] == :span
-          "<#{el.value}#{html_attributes(el.attr)}" << (!res.empty? || HTML_TAGS_WITH_BODY.include?(el.value) ? ">#{res}</#{el.value}>" : " />")
+          "<#{el.value}#{html_attributes(el.attr)}" + \
+            (!res.empty? || HTML_TAGS_WITH_BODY.include?(el.value) ? ">#{res}</#{el.value}>" : " />")
         else
-          output = ''
+          output = +''
           attr = el.attr.dup
           attr['markdown'] = '1' if markdown_attr
           output << "<#{el.value}#{html_attributes(attr)}"
           if !res.empty? && el.options[:content_model] != :block
             output << ">#{res}</#{el.value}>"
           elsif !res.empty?
-            output << ">\n#{res}"  <<  "</#{el.value}>"
+            output << ">\n#{res}" << "</#{el.value}>"
           elsif HTML_TAGS_WITH_BODY.include?(el.value)
             output << "></#{el.value}>"
           else
@@ -212,14 +221,15 @@ module Kramdown
         end
       end
 
-      def convert_xml_comment(el, opts)
-        if el.options[:category] == :block && (@stack.last.type != :html_element || @stack.last.options[:content_model] != :raw)
+      def convert_xml_comment(el, _opts)
+        if el.options[:category] == :block &&
+            (@stack.last.type != :html_element || @stack.last.options[:content_model] != :raw)
           el.value + "\n"
         else
           el.value.dup
         end
       end
-      alias :convert_xml_pi :convert_xml_comment
+      alias convert_xml_pi convert_xml_comment
 
       def convert_table(el, opts)
         opts[:alignment] = el.options[:alignment]
@@ -228,10 +238,10 @@ module Kramdown
 
       def convert_thead(el, opts)
         rows = inner(el, opts)
-        if opts[:alignment].all? {|a| a == :default}
-          "#{rows}|" << "-"*10 << "\n"
+        if opts[:alignment].all? {|a| a == :default }
+          "#{rows}|#{'-' * 10}\n"
         else
-          "#{rows}| " << opts[:alignment].map do |a|
+          "#{rows}| " + opts[:alignment].map do |a|
             case a
             when :left then ":-"
             when :right then "-:"
@@ -243,25 +253,25 @@ module Kramdown
       end
 
       def convert_tbody(el, opts)
-        res = ''
+        res = +''
         res << inner(el, opts)
-        res << '|' << '-'*10 << "\n" if opts[:next] && opts[:next].type == :tbody
+        res << '|' << '-' * 10 << "\n" if opts[:next] && opts[:next].type == :tbody
         res
       end
 
       def convert_tfoot(el, opts)
-        "|" << "="*10 << "\n#{inner(el, opts)}"
+        "|#{'=' * 10}\n#{inner(el, opts)}"
       end
 
       def convert_tr(el, opts)
-        "| " << el.children.map {|c| convert(c, opts)}.join(" | ") << " |\n"
+        "| #{el.children.map {|c| convert(c, opts) }.join(' | ')} |\n"
       end
 
       def convert_td(el, opts)
         inner(el, opts)
       end
 
-      def convert_comment(el, opts)
+      def convert_comment(el, _opts)
         if el.options[:category] == :block
           "{::comment}\n#{el.value}\n{:/}\n"
         else
@@ -269,7 +279,7 @@ module Kramdown
         end
       end
 
-      def convert_br(el, opts)
+      def convert_br(_el, _opts)
         "  \n"
       end
 
@@ -277,7 +287,7 @@ module Kramdown
         if el.attr['href'].empty?
           "[#{inner(el, opts)}]()"
         elsif el.attr['href'] =~ /^(?:http|ftp)/ || el.attr['href'].count("()") > 0
-          index = if link_el = @linkrefs.find {|c| c.attr['href'] == el.attr['href']}
+          index = if (link_el = @linkrefs.find {|c| c.attr['href'] == el.attr['href'] })
                     @linkrefs.index(link_el) + 1
                   else
                     @linkrefs << el
@@ -290,34 +300,35 @@ module Kramdown
         end
       end
 
-      def convert_img(el, opts)
+      def convert_img(el, _opts)
         alt_text = el.attr['alt'].to_s.gsub(ESCAPED_CHAR_RE) { $1 ? "\\#{$1}" : $2 }
-        if el.attr['src'].empty?
+        src = el.attr['src'].to_s
+        if src.empty?
           "![#{alt_text}]()"
         else
           title = parse_title(el.attr['title'])
-          link = if el.attr['src'].count("()") > 0
-                   "<#{el.attr['src']}>"
+          link = if src.count("()") > 0
+                   "<#{src}>"
                  else
-                   el.attr['src']
+                   src
                  end
           "![#{alt_text}](#{link}#{title})"
         end
       end
 
-      def convert_codespan(el, opts)
+      def convert_codespan(el, _opts)
         delim = (el.value.scan(/`+/).max || '') + '`'
         "#{delim}#{' ' if delim.size > 1}#{el.value}#{' ' if delim.size > 1}#{delim}"
       end
 
-      def convert_footnote(el, opts)
+      def convert_footnote(el, _opts)
         @footnotes << [el.options[:name], el.value]
         "[^#{el.options[:name]}]"
       end
 
-      def convert_raw(el, opts)
+      def convert_raw(el, _opts)
         attr = (el.options[:type] || []).join(' ')
-        attr = " type=\"#{attr}\"" if attr.length > 0
+        attr = " type=\"#{attr}\"" unless attr.empty?
         if @stack.last.type == :html_element
           el.value
         elsif el.options[:category] == :block
@@ -337,28 +348,28 @@ module Kramdown
           (opts[:next] && [:em, :strong].include?(opts[:next].type) && !ial_for_element(el) ? '{::}' : '')
       end
 
-      def convert_entity(el, opts)
+      def convert_entity(el, _opts)
         entity_to_str(el.value, el.options[:original])
       end
 
       TYPOGRAPHIC_SYMS = {
-        :mdash => '---', :ndash => '--', :hellip => '...',
-        :laquo_space => '<< ', :raquo_space => ' >>',
-        :laquo => '<<', :raquo => '>>'
+        mdash: '---', ndash: '--', hellip: '...',
+        laquo_space: '<< ', raquo_space: ' >>',
+        laquo: '<<', raquo: '>>'
       }
-      def convert_typographic_sym(el, opts)
+      def convert_typographic_sym(el, _opts)
         TYPOGRAPHIC_SYMS[el.value]
       end
 
-      def convert_smart_quote(el, opts)
+      def convert_smart_quote(el, _opts)
         el.value.to_s =~ /[rl]dquo/ ? "\"" : "'"
       end
 
-      def convert_math(el, opts)
+      def convert_math(el, _opts)
         "$$#{el.value}$$" + (el.options[:category] == :block ? "\n" : '')
       end
 
-      def convert_abbreviation(el, opts)
+      def convert_abbreviation(el, _opts)
         el.value
       end
 
@@ -371,27 +382,27 @@ module Kramdown
       end
 
       def create_link_defs
-        res = ''
-        res << "\n\n" if @linkrefs.size > 0
+        res = +''
+        res << "\n\n" unless @linkrefs.empty?
         @linkrefs.each_with_index do |el, i|
           title = parse_title(el.attr['title'])
-          res << "[#{i+1}]: #{el.attr['href']}#{title}\n"
+          res << "[#{i + 1}]: #{el.attr['href']}#{title}\n"
         end
         res
       end
 
       def create_footnote_defs
-        res = ''
+        res = +''
         @footnotes.each do |name, data|
           res << "[^#{name}]:\n"
-          res << inner(data).chomp.split(/\n/).map {|l| "    #{l}"}.join("\n") + "\n\n"
+          res << inner(data).chomp.split(/\n/).map {|l| "    #{l}" }.join("\n") + "\n\n"
         end
         res
       end
 
       def create_abbrev_defs
         return '' unless @root.options[:abbrev_defs]
-        res = ''
+        res = +''
         @root.options[:abbrev_defs].each do |name, text|
           res << "*[#{name}]: #{text}\n"
           res << ial_for_element(Element.new(:unused, nil, @root.options[:abbrev_attr][name])).to_s << "\n\n"
@@ -401,25 +412,25 @@ module Kramdown
 
       # Return the IAL containing the attributes of the element +el+.
       def ial_for_element(el)
-        res = el.attr.map do |k,v|
+        res = el.attr.map do |k, v|
           next if [:img, :a].include?(el.type) && ['href', 'src', 'alt', 'title'].include?(k)
           next if el.type == :header && k == 'id' && !v.strip.empty?
           if v.nil?
             ''
           elsif k == 'class' && !v.empty? && !v.index(/[\.#]/)
-            " " + v.split(/\s+/).map {|w| ".#{w}"}.join(" ")
+            " " + v.split(/\s+/).map {|w| ".#{w}" }.join(" ")
           elsif k == 'id' && !v.strip.empty?
             " ##{v}"
           else
-            " #{k}=\"#{v.to_s}\""
+            " #{k}=\"#{v}\""
           end
         end.compact.join('')
-        res = "toc" << (res.strip.empty? ? '' : " #{res}") if (el.type == :ul || el.type == :ol) &&
-          (el.options[:ial][:refs].include?('toc') rescue nil)
-        res = "footnotes" << (res.strip.empty? ? '' : " #{res}") if (el.type == :ul || el.type == :ol) &&
-          (el.options[:ial][:refs].include?('footnotes') rescue nil)
+        res = "toc" + (res.strip.empty? ? '' : " #{res}") if (el.type == :ul || el.type == :ol) &&
+          el.options.dig(:ial, :refs)&.include?('toc')
+        res = "footnotes" + (res.strip.empty? ? '' : " #{res}") if (el.type == :ul || el.type == :ol) &&
+          el.options.dig(:ial, :refs)&.include?('footnotes')
         if el.type == :dl && el.options[:ial] && el.options[:ial][:refs]
-          auto_ids = el.options[:ial][:refs].select {|ref| ref =~ /\Aauto_ids/}.join(" ")
+          auto_ids = el.options[:ial][:refs].select {|ref| ref.start_with?('auto_ids') }.join(" ")
           res = auto_ids << (res.strip.empty? ? '' : " #{res}") unless auto_ids.empty?
         end
         res.strip.empty? ? nil : "{:#{res}}"
